unit CachedBuffers;


{ ******************************************************************** }
{ CachedBuffers is a simple and powerful object library to have cached }
{ data reading and writing with the best performance.                  }
{                                                                      }
{ Copyright: Dmitry Mozulyov (aka Devil)                               }
{ email: softforyou@inbox.ru                                           }
{ icq: 250481638                                                       }
{ ******************************************************************** }


 
// compiler options
{$V+}{$B-}{$X+}{$T+}{$P+}{$H+}{$J-}{$MINENUMSIZE 1}
{$ifndef VER140}
  {$WARN UNSAFE_CODE OFF}
  {$WARN UNSAFE_TYPE OFF}
  {$WARN UNSAFE_CAST OFF}
{$endif}
{$O+}{$R-}{$I-}{$Q-}

interface
  uses {$ifdef MSWINDOWS}Windows,{$endif}
       Types, SysUtils;


type
  ECachedBuffer = class(Exception);
  TCachedBufferMode = (cbReader, cbWriter);
  PCustomCachedBuffer = ^TCustomCachedBuffer;
  TCachedBufferEvent = procedure(const Sender: PCustomCachedBuffer) of object;
  TCachedBufferCallback = function(Sender: PCustomCachedBuffer; Data: pointer; Size: integer): integer;

  // allocated memory information
  TCachedBufferMemory = record
    Handle: pointer;
    PreviousSize: integer;   // 0..inf = Handle..Buffer
      Buffer: pointer;
      BufferSize: integer;
      FilledSize: integer;
    AdditionalSize: integer; // bonus memory after "Buffer+BufferSize", 4kb minimum
  end;

  // internal interface to use double threading, (rarely used)
  TCachedBufferThread = class
  private
    FAutoDestroy: boolean;
    FCachedBuffer: PCustomCachedBuffer;
    FMemory: TCachedBufferMemory;
    FRunningSize: integer;
  protected
    procedure RunWrite(Data: pointer; Size: integer); virtual;
    procedure RunRead(Data: pointer; Size: integer); virtual;
    function WaitWrite: integer; virtual;
    function WaitRead: integer; virtual;
    property RunningSize: integer read FRunningSize write FRunningSize; 
  public
    constructor Create(const ACachedBuffer: PCustomCachedBuffer; const PrevPagesCount, BufferPagesCount: integer);
    destructor Destroy; override;

    procedure Run;
    function Wait: integer;
    function FullRunWait(Data: pointer; Size: integer): integer;

    property CachedBuffer: PCustomCachedBuffer read FCachedBuffer;
    property Memory: TCachedBufferMemory read FMemory;
    property AutoDestroy: boolean read FAutoDestroy write FAutoDestroy;
  end;


  // base object for cached reading/writing data
  TCustomCachedBuffer = object
  private
    FMemory: TCachedBufferMemory;
    FThread: TCachedBufferThread;
    FMode: TCachedBufferMode;
    FFinishing: boolean;

    FCallback: TCachedBufferCallback;
    FFlushCount: dword;
    FOnProgress: TCachedBufferEvent;

    function GetPosition: int64;
    procedure FlushWriter();
    procedure FlushReader();
//    procedure RunCallback(const use_thread_memory: boolean);
{    procedure SwapBuffers();
    procedure StartThread();
    procedure FinishThread(); }
  protected
    procedure Initialize(AMode: TCachedBufferMode; PrevPagesCount, BufferPagesCount: integer; ACallback: TCachedBufferCallback; AThread: TCachedBufferThread=nil);
    procedure Finalize();

    property Callback: TCachedBufferCallback read FCallback write FCallback;
    property Thread: TCachedBufferThread read FThread write FThread;
  public
    // you can use highlevel Write(TCachedBufferWriter) and Read(TCachedBufferReader) methods,
    // but to have the really maximum performance, you should use Current, Margin and Flush()

    // current pointer in cached buffer
    Current: pointer;
    // available memory size in cached buffer
    Margin: integer;

    // call Flush() when the buffer ends (small or zero Margin) to update buffer, Current and Margin
    // - don't call Flush if Finishing flag is on (which is set when
    //   callback reading/writing size is not equal fixed buffer size)
    // - you can fill some additional memory in TCachedBufferWriter (Margin < 0),
    //   the memory will be automatically moved to Buffer area after Flush time,
    //   but be attention, AdditionalSize may be 4kb only
    // - you can leave some margin memory in TCachedBufferReader before Flush calling,
    //   the memory will be available, because it's automatically moved in Previous area.
    //   negative Margin (additional memory) is ignored in TCachedBufferReader
    procedure Flush();

    // useful properties
    property Mode: TCachedBufferMode read FMode;
    property Finishing: boolean read FFinishing;
    property FlushCount: dword read FFlushCount;
    property Position: int64 read GetPosition;
    property OnProgress: TCachedBufferEvent read FOnProgress write FOnProgress;
    property Memory: TCachedBufferMemory read FMemory;
  end;


  PCachedBufferWriter = ^TCachedBufferWriter;
  TCachedBufferWriterCallback = function(const Sender: PCachedBufferWriter; const Data: pointer; const Size: integer): integer;
  
  // simple object for cached data writing
  TCachedBufferWriter = object(TCustomCachedBuffer)
  public
    procedure Initialize(const PagesCount: integer; const AWriteCallback: TCachedBufferWriterCallback; AThread: TCachedBufferThread=nil);

    // smart data writing
    procedure Write(const Buffer; const Count: integer);
  end;


  // file oriented cached buffer writer,
  // buffer size is 256kb
  TCachedFileWriter = object(TCachedBufferWriter)
  private
    FFileName: string;
    FHandle: integer;
  protected
    function InternalWriter(const Data: pointer; const Size: integer): integer;
  public
    procedure Initialize(const AFileName: string; const ADoubleBuffering: boolean=true);
    procedure Finalize();

    property FileName: string read FFileName;
  end;


  PCachedBufferReader = ^TCachedBufferReader;
  TCachedBufferReaderCallback = function(const Sender: PCachedBufferReader; const Data: pointer; const Size: integer): integer;

  // simple object for cached data reading
  TCachedBufferReader = object(TCustomCachedBuffer)
  public
    procedure Initialize(const PagesCount: integer; const AReaderCallback: TCachedBufferReaderCallback; AThread: TCachedBufferThread=nil);

    // smart data reading
    procedure Read(out Buffer; const Count: integer);
  end;


  // file oriented cached buffer reader,
  // buffer size is 256kb
  TCachedFileReader = object(TCachedBufferReader)
  private
    FFileName: string;
    FHandle: integer;
    FSize: int64;
  protected
    function InternalReader(const Data: pointer; const ASize: integer): integer;
  public
    procedure Initialize(const AFileName: string; const ADoubleBuffering: boolean=true);
    procedure Finalize();

    property FileName: string read FFileName;
    property Size: int64 read FSize;
  end;


implementation

const
  MEMORY_PAGE_SIZE = 4*1024; // 4kb
  FILE_BUFFER_SIZE = 256*1024; // 256kb

// allocate needed memory, align and fill all fields
function CachedBufferMemoryAlloc(prev_pages, BufferPages: integer): TCachedBufferMemory;
var
  Offset: integer;
begin
  Result.PreviousSize := prev_pages*MEMORY_PAGE_SIZE;
  Result.BufferSize := BufferPages*MEMORY_PAGE_SIZE;
  Result.AdditionalSize := MEMORY_PAGE_SIZE;

  with Result do
  GetMem(Handle, PreviousSize+BufferSize+AdditionalSize+{aligning}MEMORY_PAGE_SIZE);

  Offset := integer(Result.Handle) and (MEMORY_PAGE_SIZE-1);
  inc(Result.PreviousSize, MEMORY_PAGE_SIZE-Offset);
  inc(Result.AdditionalSize, Offset);
  Result.Buffer := pointer(integer(Result.Handle)+Result.PreviousSize);
end;

// free allocated memory
procedure CachedBufferMemoryFree(var Memory: TCachedBufferMemory);
begin
  if (Memory.Handle <> nil) then
  begin
    FreeMem(Memory.Handle);
    Memory.Handle := nil;
  end;
end;

// swap records
procedure CachedBufferMemorySwap(var M1, M2: TCachedBufferMemory);
var
  Buf: TCachedBufferMemory;
begin
  Buf := M1;
  M1 := M2;
  M2 := Buf;
end;

// internal additional memory copying
procedure CachedBufferMemoryCopyWriterOffset(var Dest, Src: TCachedBufferMemory; const Offset: integer);
begin
  Move(Pointer(Integer(Src.Buffer)+Src.BufferSize)^, Dest.Buffer^, Offset);
end;

// internal margin memory copying to previous area 
procedure CachedBufferMemoryCopyReaderOffset(var Dest, Src: TCachedBufferMemory; const Offset: integer);
var
  Buf: TCachedBufferMemory;

  procedure DoCopy(var D, S: TCachedBufferMemory);
  begin
    Move(Pointer(Integer(S.Buffer)+S.BufferSize-Offset)^,
         Pointer(Integer(D.Buffer)-Offset)^, Offset);
  end;
begin
  if (Offset <= Dest.PreviousSize) then
  begin
    DoCopy(Dest, Src);
  end else
  begin
    Buf := CachedBufferMemoryAlloc((Offset + (MEMORY_PAGE_SIZE-1)) div MEMORY_PAGE_SIZE,
                                   Src.BufferSize div MEMORY_PAGE_SIZE);
    DoCopy(Buf, Src);
    CachedBufferMemoryFree(Dest);
    Dest := Buf;
  end;
end;

procedure RaiseCannotReadWrite(const Mode: TCachedBufferMode; const Size: integer);
const
  READ_WRITE: array[TCachedBufferMode] of AnsiString = ('read', 'write');
begin
  raise ECachedBuffer.CreateFmt('Cannot %s %d bytes', [READ_WRITE[Mode], Size]);
end;

procedure RaiseFinishing(const Mode: TCachedBufferMode; const Action: PAnsiChar);
const
  READER_WRITER: array[TCachedBufferMode] of AnsiString = ('reader', 'writer');
begin
  raise ECachedBuffer.CreateFmt('Cannot %s, because %s is already finishing', [Action, READER_WRITER[Mode]]);
end;
  

{ TCachedBufferThread }

constructor TCachedBufferThread.Create(const ACachedBuffer: PCustomCachedBuffer;
                                const PrevPagesCount, BufferPagesCount: integer);
begin
  inherited Create;

  FCachedBuffer := ACachedBuffer;
  FMemory := CachedBufferMemoryAlloc(PrevPagesCount, BufferPagesCount);
end;

destructor TCachedBufferThread.Destroy;
begin
  if (RunningSize <> 0) then Wait();
  CachedBufferMemoryFree(FMemory);
  inherited;
end;

procedure TCachedBufferThread.RunWrite(Data: pointer; Size: integer);
begin
  RunningSize := Size;
  if (Size <= 0) then RaiseCannotReadWrite(CachedBuffer.FMode, Size);
end;

procedure TCachedBufferThread.RunRead(Data: pointer; Size: integer);
begin
  RunningSize := Size;
  if (Size <= 0) then RaiseCannotReadWrite(CachedBuffer.FMode, Size);
end;

function TCachedBufferThread.WaitWrite: integer;
begin
  Result := 0;
end;

function TCachedBufferThread.WaitRead: integer;
begin
  Result := 0;
end;

procedure TCachedBufferThread.Run;
begin
  if (RunningSize <> 0) then
  begin
    raise ECachedBuffer.Create('Thread is already running');
  end;

  if (CachedBuffer.FMode = cbWriter) then RunWrite(Memory.Buffer, Memory.FilledSize)
  else RunRead(Memory.Buffer, Memory.BufferSize);
end;

function TCachedBufferThread.Wait: integer;
begin
  if (RunningSize = 0) then
  begin
    raise ECachedBuffer.Create('Thread is already stoped');
  end;

  if (CachedBuffer.FMode = cbWriter) then
  begin
    Result := WaitWrite();
    if (Result <> RunningSize) then RaiseCannotReadWrite(cbWriter, RunningSize);
  end else
  begin
    Result := WaitRead();
    FMemory.FilledSize := Result;
  end;  
  RunningSize := 0;
end;

function TCachedBufferThread.FullRunWait(Data: pointer; Size: integer): integer;
begin
  if (RunningSize <> 0) then
  begin
    raise ECachedBuffer.Create('Thread is already running');
  end;

  if (CachedBuffer.FMode = cbWriter) then
  begin
    RunWrite(Data, Size);
    if (RunningSize = 0) then Result := Size else Result := WaitWrite();
  end else
  begin
    RunRead(Data, Size);
    if (RunningSize = 0) then Result := Size else Result := WaitRead();
  end;
  RunningSize := 0;
end;


{ TCustomCachedBuffer }


procedure TCustomCachedBuffer.Initialize(AMode: TCachedBufferMode; PrevPagesCount, BufferPagesCount: integer; ACallback: TCachedBufferCallback; AThread: TCachedBufferThread);
begin
  FMemory := CachedBufferMemoryAlloc(PrevPagesCount, BufferPagesCount);
  FThread := AThread;

  FMode := AMode;
  FFinishing := false;
  FCallback := ACallback;
  FFlushCount := 0;
  FOnProgress := nil;

  if (FMode = cbWriter) then
  begin
    Current := FMemory.Buffer;
    Margin := FMemory.BufferSize;
  end else
  // if (FMode = cbReader) then
  begin
    Current := pointer(integer(FMemory.Buffer)+FMemory.BufferSize);
    Margin := 0;
    Flush();
  end;
end;

procedure TCustomCachedBuffer.Finalize;
begin
  if (FMode = cbWriter) and (not FFinishing) then Flush();
  if (FThread <> nil) and (FThread.AutoDestroy) then FThread.Free;
  CachedBufferMemoryFree(FMemory);
end;

function TCustomCachedBuffer.GetPosition: int64;
begin
  // FFlushCount for writing and (FFlushCount - 1) for reading
  Result := (int64(FFlushCount+dword(ord(FMode))-1)*FMemory.BufferSize) + (integer(Current)-integer(FMemory.Buffer));
end;

{procedure TCustomCachedBuffer.RunCallback(const use_FThread_memory: boolean);
var
  Mem: ^TCachedBufferMemory;
  Ret: integer;

  procedure RaiseRet();
  begin
    RaiseCannotReadWrite(FMode, Ret);
  end;

begin}
(*  if (not use_FThread_memory) then Mem := @FMemory
  else Mem := @PCachedBufferFThreadMemory(FThreadMem).Memory;

  if (FMode = cbReader) then
  begin
    Ret := Callback(@Self, Mem.Buffer, Mem.BufferSize);
    Mem.FilledSize := Ret;
    if (Ret < 0) or (Ret > Mem.BufferSize) then RaiseRet();

    // finishing flag
    if (Ret <> Mem.BufferSize) then
    begin
      if (use_FThread_memory) then FThreadFinishing := true
      else FFinishing := true;
    end;
  end else
  // if (Mode = cbWriter) then
  begin
    Ret := Callback(@Self, Mem.Buffer, Mem.FilledSize);
    if (Ret <> Mem.FilledSize) then RaiseRet();
  end; *)
//end;


// difficult routine with data buffer swapping
// and run callback (in single FThread mode)
(*procedure TCustomCachedBuffer.SwapBuffers();
var
  FThreadMem: PCachedBufferFThreadMemory;
  DestMemory: ^TCachedBufferMemory;                    // cbWriter
  SrcMemory: ^TCachedBufferMemory absolute DestMemory; // cbReader
  Offset: integer;
  LastHandle, LastBuffer: pointer;
  Buf: TCachedBufferMemory;

  procedure CopyMarginMemoryToSrcPrevMemory();
  begin
    Move(Current^, pointer(integer(SrcMemory.Buffer)-Margin)^, Margin);
  end;

  function AllocPreviousMarginMemory(): TCachedBufferMemory;
  begin
    Result := CachedBufferMemoryAlloc((Margin+MEMORY_PAGE_SIZE-1) div MEMORY_PAGE_SIZE,
                                      FMemory.BufferSize div MEMORY_PAGE_SIZE);
  end;
begin
  FThreadMem := FThreadMem;
  if (FThreadMem <> nil) then DestMemory{SrcMemory} := @FThreadMem.Memory
  else DestMemory{SrcMemory} := @Self.FMemory;

  // individual margin copying (and callback in single FThread mode)
  Offset := 0;
  if (FMode = cbWriter) then
  begin
    FMemory.FilledSize := FMemory.BufferSize;
    if (Margin > 0) then
    begin
      dec(FMemory.FilledSize, Margin);

      // finishing flag
      if (FThreadMem <> nil) then FThreadFinishing := true;
      FFinishing := true;
    end;

    if (FThreadMem = nil) then RunCallback(false);

    // additional memory is copied to start of the buffer
    if (Margin < 0) then
    begin
      Offset := -Margin;
      Move(pointer(integer(FMemory.Buffer)+FMemory.BufferSize)^, DestMemory.Buffer^, Offset);
    end;
  end else
  // if (Mode = cbReader) then
  begin
    if (Margin < 0) then Margin := 0;
    Offset := -Margin;

    // too big Margin
    if (Margin > SrcMemory.PreviousSize) then
    begin
      LastHandle := SrcMemory.Handle;

      if (FThreadMem = nil) then
      begin
        FMemory := AllocPreviousMarginMemory();
        CopyMarginMemoryToSrcPrevMemory();
        FreeMem(LastHandle);
        RunCallback(false);
      end else
      // grow source FThread memory buffer
      begin
        LastBuffer := SrcMemory.Buffer;
        SrcMemory^ := AllocPreviousMarginMemory();
        CopyMarginMemoryToSrcPrevMemory();
        Move(LastBuffer^, SrcMemory.Buffer^, SrcMemory.BufferSize);
        FreeMem(LastHandle);
      end;
    end else
    // normal Margin
    begin
      if (Margin > 0) then CopyMarginMemoryToSrcPrevMemory();
      if (FThreadMem = nil) then RunCallback(false);
    end;
  end;

  // swap memory buffers
  if (FThreadMem <> nil) then
  begin
    Buf := DestMemory^;
    DestMemory^ := FMemory;
    FMemory := Buf;

    if (FThreadMem.State = STATE_DONE) then  FThreadMem.State := STATE_PROCESSING;
  end;

  // Current, Margin
  Current := FMemory.Buffer;
  if (FMode = cbWriter) then Margin := FMemory.BufferSize else Margin := FMemory.FilledSize;
  inc(integer(Current), Offset);
  dec(Margin, Offset);
end; *)



procedure TCustomCachedBuffer.FlushWriter();
begin
  FMemory.FilledSize := FMemory.BufferSize;

  // if last
  if (Margin > 0) then
  begin
    dec(FMemory.FilledSize, Margin);
    FFinishing := true;
    Margin := 0;
    if (FMemory.FilledSize = 0) then exit;
  end;

  // different algorithms: standard and double FThreading
  if (FThread <> nil) then
  begin
    if (FFlushCount <> 1) then FThread.Wait();

    // swap
    CachedBufferMemorySwap(Self.FMemory, FThread.FMemory);

    // copy margin
    if (Margin < 0) then CachedBufferMemoryCopyWriterOffset(Self.FMemory, FThread.FMemory, -Margin);

    // thread running
    Thread.Run();
  end else
  // simple mode
  begin
    // run Callback
    if (Self.FMemory.FilledSize <> Callback(@Self, Self.FMemory.Buffer, Self.FMemory.FilledSize)) then
    RaiseCannotReadWrite(cbWriter, Self.FMemory.FilledSize);

    // copy margin (if needed)
    if (Margin < 0) then
    CachedBufferMemoryCopyWriterOffset(Self.FMemory, Self.FMemory, -Margin);
  end;

  // fill Current and Margin
  Current := Pointer(Integer(FMemory.Buffer) - Margin{<=0});
  Margin := FMemory.BufferSize + Margin{<=0};
end;

procedure TCustomCachedBuffer.FlushReader();
begin
  if (Margin < 0) then Margin := 0;

  // different algorithms: standard and double FThreading
  if (FThread <> nil) then
  begin
    if (FFlushCount = 1) then FThread.Run();
    if (FThread.FMemory.BufferSize <> FThread.Wait()) then FFinishing := true;

    CachedBufferMemorySwap(Self.FMemory, FThread.FMemory);
    if (Margin > 0) then CachedBufferMemoryCopyReaderOffset(Self.FMemory, FThread.FMemory, Margin);

    if (not FFinishing) then FThread.Run();
  end else
  // simple mode
  begin
    if (Margin > 0) then CachedBufferMemoryCopyReaderOffset(Self.FMemory, Self.FMemory, Margin);
    Self.FMemory.FilledSize := Callback(@Self, Self.FMemory.Buffer, Self.FMemory.BufferSize);
    if (Self.FMemory.BufferSize <> Self.FMemory.FilledSize) then FFinishing := true;
  end;

  // fill Current and Margin
  Current := Pointer(Integer(FMemory.Buffer) - Margin{>=0});
  Margin := FMemory.FilledSize + Margin{>=0};
end;

procedure TCustomCachedBuffer.Flush();
begin
  if (FFinishing) then RaiseFinishing(FMode, 'flush');
  inc(FFlushCount);

  if (FMode = cbWriter) then FlushWriter()
  else FlushReader();

  // event
  if (assigned(FOnProgress)) then FOnProgress(@Self);
end;




{ TCachedBufferWriter }

procedure TCachedBufferWriter.Initialize(const PagesCount: integer; const AWriteCallback: TCachedBufferWriterCallback; AThread: TCachedBufferThread=nil);
begin
  inherited Initialize(cbWriter, 0, PagesCount, TCachedBufferCallback(AWriteCallback), AThread);
end;


// smart data writing
// todo asm ?
procedure TCachedBufferWriter.Write(const Buffer; const Count: integer);
var
  Data: pointer;
  Size: integer;
  C, S: integer;  
begin
  if (FFinishing) then RaiseFinishing(cbWriter, 'write data');

  if (Count <= Margin) then
  begin
    Move(Buffer, Current^, Count);
    inc(integer(Current), Count);
    dec(Margin, Count);
    if (Margin = 0) then Flush();
  end else
  // if (Count > Margin) then
  begin
    Data := @Buffer;
    Size := Count;

    // write to buffer's end and Flush
    if (Current <> FMemory.Buffer) or (FFlushCount = 0) then
    begin
      if (Margin <> 0) then
      begin
        Move(Data^, Current^, Margin);

        inc(integer(Current), Margin);
        inc(integer(Data), Margin);
        dec(Size, Margin);
        Margin := 0;
      end;

      Flush();
    end;

    // if written data is too large, we can call writer
    // without memory buffer using
    if (Size >= FMemory.BufferSize) then
    begin
      if (FThread <> nil) then FThread.Wait();
      C := Size div FMemory.BufferSize;
      S := C * FMemory.BufferSize;

      if (FThread <> nil) then
      begin
        FThread.FullRunWait(Data, S);
      end else
      begin
        if (S <> Callback(@Self, Data, S)) then RaiseCannotReadWrite(cbWriter, S);
      end;

      Inc(Integer(Data), S);
      Dec(Size, S);
      inc(FFlushCount, C);
      if (assigned(FOnProgress)) then FOnProgress(@Self);
    end;

    // last Data bytes
    if (Size <> 0) then
    begin
      Move(Data^, Current^, Size);
      inc(integer(Current), Size);
      dec(Margin, Size);
    end;
  end;
end;



{$ifdef MSWINDOWS}
type
  TWinFileBufferFThread = class(TCachedBufferThread)
  private
    FFile: THandle;
    FNoBuffSectorSize: integer;
    F: packed record
    case Integer of
      0: (Struct: Windows._OVERLAPPED);
      1: (__: int64; Offset: int64; Event: THandle);
    end;

    function OverlappedWait: integer;
  protected
    procedure RunWrite(Data: pointer; Size: integer); override;
    procedure RunRead(Data: pointer; Size: integer); override;
    function WaitWrite: integer; override;
    function WaitRead: integer; override;

    property NoBuffSectorSize: integer read FNoBuffSectorSize;
  public
    constructor Create(ACachedBuffer: PCustomCachedBuffer; AFile: THandle; AReader: boolean; ANoBuffSectorSize: integer);
    destructor Destroy; override;
  end;

{ TWinFileBufferFThread }

constructor TWinFileBufferFThread.Create(ACachedBuffer: PCustomCachedBuffer; AFile: THandle; AReader: boolean; ANoBuffSectorSize: integer);
begin
  FFile := AFile;
  F.Event := Windows.CreateEvent(nil, True, False, nil); // todo delete ?
  FNoBuffSectorSize := ANoBuffSectorSize;  
  inherited Create(ACachedBuffer, ord(AReader), FILE_BUFFER_SIZE div MEMORY_PAGE_SIZE);
end;

destructor TWinFileBufferFThread.Destroy;
begin
  inherited;
  CloseHandle(F.Event); // todo delete ?
end;

function TWinFileBufferFThread.OverlappedWait: integer;
begin
  if (not Windows.GetOverlappedResult(FFile, F.Struct, dword(Result), TRUE)) then RaiseLastOSError();
  F.Offset := F.Offset + Result;
end;

procedure TWinFileBufferFThread.RunWrite(Data: pointer; Size: integer);
var
  fake: dword;
  Count, i: integer;
begin
  inherited;

  if (NoBuffSectorSize = 0) or ((integer(Data) and (NoBuffSectorSize-1))=0) then
  begin
    Windows.WriteFile(FFile, Data^, Size, fake, @F.Struct);
  end else
  // difficult non-buffered case!
  begin
    Count := (Size+Memory.BufferSize-1) div Memory.BufferSize;

    for i := 0 to Count-2 do
    begin
      Move(Data^, Memory.Buffer^, Memory.BufferSize);
      Windows.WriteFile(FFile, Memory.Buffer^, Memory.BufferSize, fake, @F.Struct);
      if (OverlappedWait() <> Memory.BufferSize) then RaiseLastOSError();

      inc(integer(Data), Memory.BufferSize);
      dec(Size, Memory.BufferSize);
    end;

    Move(Data^, Memory.Buffer^, Size);
    Windows.WriteFile(FFile, Memory.Buffer^, Size, fake, @F.Struct);
  end;
end;

procedure TWinFileBufferFThread.RunRead(Data: pointer; Size: integer);
var
  fake: dword;
  Count, i: integer;

  procedure RaiseReading;
  begin
    RaiseCannotReadWrite(cbReader, Count*Memory.BufferSize + (Size mod Memory.BufferSize));
  end;
begin
  inherited;

  if (NoBuffSectorSize = 0) or ((integer(Data) and (NoBuffSectorSize-1))=0) then
  begin
    Windows.ReadFile(FFile, Data^, Size, fake, @F.Struct);
  end else
  // difficult non-buffered case!
  begin
    Count := (Size+Memory.BufferSize-1) div Memory.BufferSize;

    for i := 0 to Count-2 do
    begin
      Windows.ReadFile(FFile, Memory.Buffer^, Memory.BufferSize, fake, @F.Struct);
      if (OverlappedWait() <> Memory.BufferSize) then RaiseReading;
      Move(Memory.Buffer^, Data^, Memory.BufferSize);

      inc(integer(Data), Memory.BufferSize);
      dec(Size, Memory.BufferSize);
    end;

    Windows.ReadFile(FFile, Memory.Buffer^, Size, fake, @F.Struct);
    if (OverlappedWait() <> Memory.BufferSize) then RaiseReading;
    Move(Memory.Buffer^, Data^, Size);

    RunningSize := 0;
  end;
end;

function TWinFileBufferFThread.WaitRead: integer;
begin
  Result := OverlappedWait();
end;

function TWinFileBufferFThread.WaitWrite: integer;
begin
  Result := OverlappedWait();
  if (Result <> RunningSize) then RaiseLastOSError();
end;
{ <-- TWinFileBufferFThread }


function Windows7_higher: boolean;
begin
  Result := false;

  case Win32Platform of
    0..VER_PLATFORM_WIN32_WINDOWS: ;
    VER_PLATFORM_WIN32_NT:
      case Win32MajorVersion of
        0..5: ;
        6: Result := (Win32MinorVersion >= 1);
      else
        Result := true;
      end;
  else
    Result := true;
  end;
end;

function GetSectorSize(const FileName: string): integer;
var
  buf: dword;
begin
  GetDiskFreeSpace(PChar(Copy(ExpandFileName(FileName), 1, 3)), buf, dword(Result), buf, buf);
end;
{$endif}



{ TCachedFileWriter }

{$ifdef MSWINDOWS}
procedure TCachedFileWriter.Initialize(const AFileName: string; const ADoubleBuffering: boolean=true);
var
  Attributes: integer;
  AThread: TWinFileBufferFThread;
begin
  AThread := nil;  
  FFileName := AFileName;

  Attributes := FILE_FLAG_SEQUENTIAL_SCAN or integer(FILE_FLAG_WRITE_THROUGH);
  if (ADoubleBuffering) then Attributes := Attributes or FILE_FLAG_OVERLAPPED;
  FHandle := Integer(CreateFile(PChar(AFileName), GENERIC_WRITE, 0, nil, CREATE_ALWAYS, Attributes, 0));
  if (FHandle < 0) then raise ECachedBuffer.CreateFmt('Cannot create file:'#13'%s', [AFileName]);

  if (ADoubleBuffering) then
  begin
    AThread := TWinFileBufferFThread.Create(@Self, FHandle, false, 0);
    AThread.AutoDestroy := true;
  end;
  inherited Initialize(FILE_BUFFER_SIZE div MEMORY_PAGE_SIZE, TCachedBufferWriterCallback(@TCachedFileWriter.InternalWriter), AThread);
end;
{$else}
procedure TCachedFileWriter.Initialize(const AFileName: string; const ADoubleBuffering: boolean=true);
begin
  FFileName := AFileName;
  FHandle := FileCreate(AFileName);
  if (FHandle < 0) then raise ECachedBuffer.CreateFmt('Cannot create file:'#13'%s', [AFileName]);
  inherited Initialize(FILE_BUFFER_SIZE div MEMORY_PAGE_SIZE, TCachedBufferWriterCallback(@TCachedFileWriter.InternalWriter));
end;
{$endif}

procedure TCachedFileWriter.Finalize();
begin
  inherited; 
  if (FHandle >= 0) then FileClose(FHandle);
  FFileName := '';
end;

function TCachedFileWriter.InternalWriter(const Data: pointer; const Size: integer): integer;
begin
  Result := FileWrite(FHandle, Data^, Size);
  if (Result <> Size) then RaiseLastOSError;
end;




{ TCachedBufferReader }

procedure TCachedBufferReader.Initialize(const PagesCount: integer; const AReaderCallback: TCachedBufferReaderCallback; AThread: TCachedBufferThread=nil);
begin
  inherited Initialize(cbReader, 1, PagesCount, TCachedBufferCallback(AReaderCallback), AThread);
end;

// smart data reading
// todo asm ?
procedure TCachedBufferReader.Read(out Buffer; const Count: integer);
var
  Data: pointer;
  Size: integer;
  C, S: integer;  

  procedure RaiseFinishedReading();
  begin
    RaiseFinishing(cbReader, 'read data');
  end;

begin
  if (Margin >= Count) then
  begin
    Move(Current^, Buffer, Count);
    inc(integer(Current), Count);
    dec(Margin, Count);
    if (Margin = 0) and (not FFinishing) then Flush();
  end else
  // Margin < Count, need to read some full buffers
  begin
    if (FFinishing) then RaiseFinishedReading();
    Data := @Buffer;
    Size := Count;

    // read Margin
    if (Margin <> 0) then
    begin
      Move(Current^, Data^, Margin);
      inc(integer(Data), Margin);
      inc(integer(Current), Margin);
      dec(Size, Margin);
      Margin := 0;
    end;

    // if read data is too large, we can call reader callback
    // without memory buffer using
    if (Size >= FMemory.BufferSize) then
    begin
      if (FThread <> nil) then
      begin
        if (FMemory.BufferSize <> FThread.Wait()) then RaiseFinishedReading();
        inc(FFlushCount);
        Move(FThread.FMemory.Buffer^, Data^, FMemory.BufferSize);
        Inc(Integer(Data), FMemory.BufferSize);
        Dec(Size, FMemory.BufferSize);
        if (assigned(FOnProgress)) then FOnProgress(@Self);
      end;

      if (Size >= FMemory.BufferSize) then
      begin
        C := Size div FMemory.BufferSize;
        S := C * FMemory.BufferSize;

        if (FThread <> nil) then
        begin
          if (S <> FThread.FullRunWait(Data, S)) then RaiseFinishedReading();
        end else
        begin
          if (S <> Callback(@Self, Data, S)) then RaiseFinishedReading();
        end;

        Inc(Integer(Data), S);
        Dec(Size, S);
        inc(FFlushCount, C);
        if (assigned(FOnProgress)) then FOnProgress(@Self);
      end;

      if (Thread <> nil) then Thread.Run();
    end;

    // last Data bytes
    if (Size <> 0) then
    begin
      Flush();
      if (Margin < Size) then RaiseFinishedReading();

      Move(Current^, Data^, Size);
      inc(integer(Current), Size);
      dec(Margin, Size);
    end;    
  end;
end;




{ TCachedFileReader }

{$ifdef MSWINDOWS}
procedure TCachedFileReader.Initialize(const AFileName: string; const ADoubleBuffering: boolean=true);
var
  Attributes: integer;
  NoBuffSectorSize: integer;
  AThread: TWinFileBufferFThread;
begin
  AThread := nil;
  Attributes := FILE_FLAG_SEQUENTIAL_SCAN;
  NoBuffSectorSize := 0;  
  if (ADoubleBuffering) then
  begin
    Attributes := Attributes or FILE_FLAG_OVERLAPPED;

    if (not Windows7_higher) then
    begin
      NoBuffSectorSize := GetSectorSize(AFileName);
      if (NoBuffSectorSize <> 0) then Attributes := Attributes or FILE_FLAG_NO_BUFFERING;
    end;
  end;  

  FFileName := AFileName;
  FHandle := Integer(CreateFile(PChar(AFileName), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, Attributes, 0));
  if (FHandle < 0) then raise ECachedBuffer.CreateFmt('Cannot read file:'#13'%s', [AFileName]);
  pdword(@FSize)^ := Windows.GetFileSize(FHandle, pdword(integer(@FSize)+sizeof(dword)));

  if (ADoubleBuffering) then
  begin
    AThread := TWinFileBufferFThread.Create(@Self, FHandle, true, NoBuffSectorSize);
    AThread.AutoDestroy := true;
  end;
  inherited Initialize(FILE_BUFFER_SIZE div MEMORY_PAGE_SIZE, TCachedBufferReaderCallback(@TCachedFileReader.InternalReader), AThread);
end;
{$else}
procedure TCachedFileReader.Initialize(const AFileName: string; const ADoubleBuffering: boolean=true);
begin
  FFileName := AFileName;
  FHandle := FileOpen(FileName, fmOpenRead or fmShareDenyNone);
  if (FHandle < 0) then raise ECachedBuffer.CreateFmt('Cannot read file:'#13'%s', [AFileName]);
  FSize := FileSeek(FHandle, int64(0), FILE_END);
  FileSeek(FHandle, 0, FILE_BEGIN);
  inherited Initialize(FILE_BUFFER_SIZE div MEMORY_PAGE_SIZE, TCachedBufferReaderCallback(@TCachedFileReader.InternalReader));
end;
{$endif}


procedure TCachedFileReader.Finalize();
begin
  inherited;
  if (FHandle >= 0) then FileClose(FHandle);
  FFileName := '';
end;

function TCachedFileReader.InternalReader(const Data: pointer; const ASize: integer): integer;
begin
  Result := FileRead(FHandle, Data^, ASize);
  if Result = -1 then Result := 0;
end;







end.
